// waifusearch, the IB searcher
// ============================
// -This software is designed to search an imageboard's text.

// -Filename: main.cpp
// -The program's entry point, and implementation details.

#include <algorithm>
#include <chrono>
#include <cstdint>
#include <filesystem>
#include <fstream>
#include <functional>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <map>
#include <numeric>
#include <set>
#include <sstream>
#include <string>
#include <utility>
#include <vector>

#if __has_include(<execution>)  // RPi will only run single-threaded for now
#include <execution>
using std::execution::par_unseq;
#endif

#include "rw_text_utils.hpp"

// alias'
namespace fs    = std::filesystem;
using file_path = fs::path;
using dir_iter  = fs::directory_iterator;
using time_pt   = std::chrono::time_point<std::chrono::steady_clock>;
//
using Pair_ints     = std::pair<uint32_t, uint32_t>;
using Set_pair_ints = std::set<Pair_ints>;
using Vec_pair_ints = std::vector<Pair_ints>;
//
using Map_words_2_posts = std::map<std::string, Set_pair_ints>;
using Vec_words_2_posts = std::vector<std::pair<std::string, Vec_pair_ints>>;
using Results_w_offset  = std::vector<std::pair<Pair_ints, long>>;

// global vars
// -user's: (feel free to tweak these with some caution, Anon)
const bool        make_crosslinks{true};
const bool        show_timing{false};
const bool        show_terms{false};
const bool        write_files{true};
const file_path   jsons_dir{"./all_jsons/"};
const std::string board_nm{
    "robowaifu"};  // you'll also need the proper JSON files collection too, ofc
//
// -system's:
std::chrono::steady_clock  clk_time{};
std::vector<time_pt>       time_spans{};
std::map<uint32_t, string> thrd_subjs{};
std::map<uint32_t, string> all_msgs{};
std::string                search{};
std::vector<std::string>   unfounds{};
bool                       have_results{};

// usings
using std::accumulate;
using std::back_inserter;
using std::cbegin;
using std::cend;
using std::cin;
using std::cout;
using std::for_each;
using std::getline;
using std::left;
using std::make_pair;
using std::map;
using std::ofstream;
using std::out_of_range;
using std::pair;
using std::set;
using std::set_intersection;
using std::setw;
using std::stable_sort;
using std::string;
using std::stringstream;
using std::to_string;
using std::vector;
using std::chrono::duration_cast;
using std::chrono::microseconds;
using std::chrono::milliseconds;
using std::chrono::steady_clock;

/**-----------------------------------------------------------------------------
 @brief    Display the iterated, formatted, output of a search bin's results to
 console, noting the character offset for an ordered (IE, exact match) result,
 if present. The display includes the assembled URI link (whether crosslink or
 hyperlink) for each post.

 @param    vtp_o [IN] The search bin to output.
*/
void disp_results(const Results_w_offset& vtp_o) {
  //
  const uint32_t pad_len{32};
  cout << rw::make_pad_str("\n   THREAD SUBJECT", pad_len)
       << "        POST LINK\n";

  uint32_t prev_t{0};

  for (const auto& tp_o : vtp_o) {
    const auto [tp, offset]{tp_o};
    const auto [t, p]{tp};

    string subj{};
    if (t != prev_t)  // display subject for each new thread in list
      subj = rw::make_pad_str(thrd_subjs.at(t), pad_len);
    else  // no need to repeat the same thread subject, just 'ditto' it
      subj = rw::make_pad_str("   \" ", pad_len);

    // DESIGN: use C++20 std::format when available

    string   uri{};
    uint32_t uri_pad_len{};
    if (make_crosslinks) {
      uri         = "   >>" + to_string(p);
      uri_pad_len = 13;
    } else {  // use a general hyperlink instead
      uri = {"  -> https://alogs.theguntretort.com/" + board_nm + "/res/" +
             to_string(t) + ".html#" + to_string(p)};
      uri_pad_len = 68;
    }

    // the actual output statement
    cout << subj
         << rw::make_pad_str(uri, uri_pad_len)
         // display the character offset of an ordered result, if present
         << ((offset >= 0) ? "@ch: ~" + to_string(offset + 1) : "") << '\n';

    prev_t = t;
  }
}

/**-----------------------------------------------------------------------------
 @brief    Front end to marshal formatted display of 'ordered' & 'unordered'
 search bins out to console, noting total results count.

 @param    ordered [IN] The ordered bin.
 @param    unordered [IN] The unordered bin.
 @param    vtp_sz [IN] The total results count.
*/
void disp_head_foot(const Results_w_offset& ordered,
                    const Results_w_offset& unordered,
                    const size_t            vtp_sz) {
  const auto ord_sz   = ordered.size();
  const auto unord_sz = unordered.size();

  if (ord_sz > 0) {
    cout << "\n                  ORDERED:\n                  ========";
    disp_results(ordered);
  }

  if (unord_sz > 0) {
    cout << "\n                  UN-ORDERED:\n                  "
            "===========";
    disp_results(unordered);
  }

  cout << "\n' " << search << " '  [" << ord_sz << " : " << unord_sz
       << "] = " << vtp_sz << " results\n";
}

/**-----------------------------------------------------------------------------
 @brief    Search the intersection vector 'vtp' for exact, ordered matches with
 the user's search terms from the text collection of all messages. Exact
 orderings are stored into one container bin, and inexact into another. They are
 [both] then further processed.

 @param    vtp [IN] The vector of intersected thread_id+post_id pairs to search.
*/
void search_ordered(const Vec_pair_ints& vtp) {
  //
  const auto bgn_ord = clk_time.now();

  Results_w_offset ordered{}, unordered{};

  for (const auto& tp : vtp) {
    const auto [t, p]{tp};
    const auto msg = all_msgs.at(p);  // read in this post's message text

    // search message for an exact, ordered match with user's search terms
    const auto it =
        std::search(cbegin(msg), cend(msg),
                    std::boyer_moore_searcher(cbegin(search), cend(search)));

    if (it != cend(msg)) {  // exact order match found
      const auto char_offset = it - cbegin(msg);
      ordered.emplace_back(make_pair(tp, char_offset));

    } else
      unordered.emplace_back(make_pair(tp, -1));
  }

  time_spans.emplace_back(clk_time.now() - bgn_ord);  // time_spans[5]: ord

  disp_head_foot(ordered, unordered, vtp.size());
}

//----------------------------------
// DESIGN:
// -Later, extend this approach to start doing near-miss spellings
// (lev_dist?), and synonyms (for vague recollections, etc.)
//

/**-----------------------------------------------------------------------------
 @brief    Intersect all the thread_id+post_id pairs in the vector
 'vw_vtp_valids' to find the complete intersection of all posts each containing
 all words in the vector. This intersection vector is then further processed.

 @param    vw_vtp_valids [IN] The vector of thread_id+post_id pairs to intersect
*/
void find_intersects(const Vec_words_2_posts& vw_vtp_valids) {
  const auto bgn_int = clk_time.now();

  // load the first (shortest) vec to 'prime-the-pump'
  auto [w, vtp] = vw_vtp_valids[0];

  const auto low_count = vtp.size();
  const auto terms_cnt = vw_vtp_valids.size();

  if (terms_cnt > 1) {  // have multiple terms, therefore perform intersections
    Vec_pair_ints tmp_vtp{};

    // note: indexing starts at 1, 0th already loaded above
    for (size_t i{1}; i < terms_cnt; ++i) {
      // short-circuit, we already have the max intersections usable
      if (i == low_count)
        break;

      const auto& [w, next_vtp] = vw_vtp_valids[i];

#if __cpp_lib_parallel_algorithm >= 201603
      set_intersection(par_unseq, cbegin(vtp), cend(vtp),
#else
      set_intersection(cbegin(vtp), cend(vtp),
#endif
                       cbegin(next_vtp), cend(next_vtp),
                       back_inserter(tmp_vtp));

      // short-circuit, we already have the earliest intersections matching
      if (vtp == tmp_vtp)
        break;

      vtp = tmp_vtp;  // copy current result for next intersection
      tmp_vtp.clear();
    }
  }
  // Note: the case of a single term was set during 'prime-the-pump' statement

  time_spans.emplace_back(clk_time.now() - bgn_int);  // time_spans[4]: int

  search_ordered(vtp);
}

/**-----------------------------------------------------------------------------
 @brief    Display located terms 'vw_vtp_valids' counts, formatted, to console.

 @param    vw_vtp_valids [IN] The vector of terms + counts to display.
 @param    pad_len [IN] A number specifying the display string lengths.
*/
void disp_terms(const Vec_words_2_posts& vw_vtp_valids,
                const uint32_t           pad_len) {
  //
  for (const auto& valid : vw_vtp_valids) {
    const auto& [term, tp]{valid};
    cout << " term:  " << rw::make_pad_str(term, pad_len)
         << "  count:  " << tp.size() << '\n';
  }

  cout << "# terms found:  " << vw_vtp_valids.size();
}

/**-----------------------------------------------------------------------------
 @brief    Sort the vector 'io_vw_vtp_counts' by post counts, ascending is
 default.

 @param    io_vw_vtp_counts [IN/OUT] The target vector to sort.
 @param    descending [IN] A flag determining sorting mode.
*/
void sort_counts(Vec_words_2_posts& io_vw_vtp_counts,
                 const bool         descending = false) {
  const auto bgn_srt = clk_time.now();

#if __cpp_lib_parallel_algorithm >= 201603
  stable_sort(par_unseq, begin(io_vw_vtp_counts), end(io_vw_vtp_counts),
#else
  stable_sort(begin(io_vw_vtp_counts), end(io_vw_vtp_counts),
#endif
              [descending](const auto& a, const auto& b) {
                const auto& [wa, tp_a]{a};
                const auto& [wb, tp_b]{b};

                if (descending)
                  return tp_a.size() > tp_b.size();
                else
                  return tp_a.size() < tp_b.size();
              });

  time_spans.emplace_back(clk_time.now() - bgn_srt);  // time_spans[3]: srt
}

/**-----------------------------------------------------------------------------
 @brief    Copy the word@post map 'w_tp's contents into vector 'io_vw_vtp'.

 @param    w_tp [IN] The source map to copy.
 @param    io_vw_vtp [IN/OUT] The target vector to populate.
*/
void cp_map_2_vec(const Map_words_2_posts& w_tp, Vec_words_2_posts& io_vw_vtp) {
  const auto bgn_cpy = clk_time.now();

#if __cpp_lib_parallel_algorithm >= 201603
  for_each(par_unseq, cbegin(w_tp), cend(w_tp), [&](const auto& e) {
#else
  for_each(cbegin(w_tp), cend(w_tp), [&](const auto& e) {
#endif
    const auto& [w, stp]{e};
    io_vw_vtp.emplace_back(make_pair(w, Vec_pair_ints{cbegin(stp), cend(stp)}));
  });

  time_spans.emplace_back(clk_time.now() - bgn_cpy);  // time_spans[2]: cpy
}

/**-----------------------------------------------------------------------------
 @brief    Search the word@post map 'w_tp' for every matching term sourced from
 the vector of strings 'terms'. Any results are then copied into other
 containers (map->map->vector), sorted, and further processed.

 @param    w_tp [IN] The map to search.
 @param    terms [IN] The vector of strings to locate.
*/
void locate_terms(const Map_words_2_posts&        w_tp,
                  const std::vector<std::string>& terms) {
  const auto bgn_loc = clk_time.now();

  const uint32_t    pad_len{27};
  Map_words_2_posts uniques{};

  // attempt a term locate inside the w_tp and if found, copy it's thrd+post set
#if __cpp_lib_parallel_algorithm >= 201603
  for_each(par_unseq, cbegin(terms), cend(terms), [&](const auto& term) {
#else
  for_each(cbegin(terms), cend(terms), [&](const auto& term) {
#endif
    try {
      // ****
      // NOTE: This 1 statement constitutes the bulk of the algorithm's workload
      uniques[term] = w_tp.at(term);  // attempt to locate & thence to copy data
      // ****
    } catch (const out_of_range& e) {
      unfounds.emplace_back(rw::make_pad_str(term, pad_len));  // global
    }
  });

  time_spans.emplace_back(clk_time.now() - bgn_loc);  // time_spans[1]: loc

  if (! uniques.empty()) {  // did we find anything?
    Vec_words_2_posts vw_vtp_valids{};

    cp_map_2_vec(uniques, vw_vtp_valids);
    sort_counts(vw_vtp_valids);  // sort terms by post counts, low to high

    if (show_terms)
      disp_terms(vw_vtp_valids, pad_len);

    find_intersects(vw_vtp_valids);

    have_results = true;  // global
  } else
    have_results = false;
}

/**-----------------------------------------------------------------------------
 @brief    Take the vector of strings 'words' and assemble each one,
 space-separated, into a new single string.

 @param    words [IN] The vector of strings to assemble.

 @returns  The assembled string.
*/
std::string assemble_words(const std::vector<std::string>& words) {
  std::string res{};

  for (const auto& word : words)
    res += (word + ' ');

  rw::trim_r(res, ' ');
  return res;
}

/**-----------------------------------------------------------------------------
 @brief    Perform an over-reaching character cleanup of the 'io_str' string to
 partially try dealing with all the unexpectedly random text not uncommon in IB
 post messages. In addition, all alphabetic chars are lower-cased here.

 @param    io_str [IN/OUT] The string to clean.
*/
void scrub_chars(std::string& io_str) {
  const vector<pair<char, char>> bad_goods{
      {'/', ' '}, {'.', ' '}, {'!', ' '},  {'?', ' '},  {'&', ' '},  {'^', ' '},
      {'$', ' '}, {'%', ' '}, {'#', ' '},  {'@', ' '},  {'*', ' '},  {'~', ' '},
      {'"', ' '}, {'=', ' '}, {'\0', ' '}, {'\n', ' '}, {'(', ' '},  {')', ' '},
      {'[', ' '}, {']', ' '}, {'<', ' '},  {'>', ' '},  {'{', ' '},  {'}', ' '},
      {',', ' '}, {':', ' '}, {'\\', ' '}, {';', ' '},  {'\t', ' '}, {'-', ' '},
      {'_', ' '}, {'`', ' '}, {'|', ' '}};

  for (auto [bad, good] : bad_goods)
    rw::repl_bad_char(io_str, bad, good);

  rw::to_lower(io_str);
}

// NOTES:

// DESIGN: Other than the legitimate single-letters 'a' & 'I', (and slang 'U'
// & 'Y', and the programming language 'C'), all the letters should be zapped if
// they are coming up as individual words, similar to the numbers (though
// numbers can at least actually have legit context. the letters thing is
// probably more of an artifact of our quick & dirty overfiltering approach
// here).

// DESIGN: Rework to create and use a 'rm_all_nums()' but guard it with a test
// first against a list of specific numbers like years, 9001, etc. to keep.

/**-----------------------------------------------------------------------------
 @brief    Clear the 'io_str' string if it matches unexpectedly random text not
 uncommon in IB post messages.

 @param    io_str [IN/OUT] The string potentially to clear.
*/
void scrub_string(std::string& io_str) {
  const vector<string> bad_strings{
      "www", "n",  "v", "'", "b",  "w", "o",   "h",    "t",    "s", "e",
      "x",   "fl", "+", "m", "fo", "r", "p",   "d",    "f",    "g", "y",
      "nn",  "u",  "–", "z", "q",  "j", "iii", "apos", "quot", "0", "1",
      "2",   "3",  "4", "5", "6",  "7", "8",   "9"};

  for (const auto& bad_string : bad_strings)
    if (io_str == bad_string)
      io_str.clear();
}

/**-----------------------------------------------------------------------------
 @brief    Wrapper to trim the 'io_str' string of any target chars (currently
 <spc> & <'>) .

 @param    io_str [IN/OUT] The string to clean.
*/
void trim_chars(std::string& io_str) {
  rw::trim_all(io_str, ' ');
  rw::trim_all(io_str, '\'');
}

/**-----------------------------------------------------------------------------
 @brief    Wrapper to remove unwanted chars from the 'io_str' string, and to
 consolidate consecutive chars (currently <spc> & <'>) down to just one per.

 @param    io_str [IN/OUT] The string to clean.
*/
void scrub_rm_dupes(std::string& io_str) {
  scrub_chars(io_str);
  rw::rm_dupes(io_str, ' ');
  rw::rm_dupes(io_str, '\'');
}

/**-----------------------------------------------------------------------------
 @brief    Parse the input string 'text' into a vector of cleaned strings.

 @param    text [IN] The string to parse.

 @returns  The vector of strings after parsing.
*/
std::vector<std::string> scrub_n_extract(const std::string& text) {
  string temp{text};
  scrub_rm_dupes(temp);

  // parse each message into a vector<string>
  stringstream   message{temp};
  vector<string> words{};
  //
  while (message >> temp) {  // parses on whitespace
    trim_chars(temp);
    scrub_string(temp);

    if (! temp.empty())
      words.push_back(temp);
  }

  return words;
}

/**-----------------------------------------------------------------------------
 @brief    Iterate the collection of unfound terms and display them to console.
*/
void disp_unfounds() {
  if (unfounds.size() > 0) {  // global var
    cout << '\n';
    for (const auto& unfound : unfounds)
      cout << " term:  " << unfound << "  ** not found\n";

    unfounds.clear();
  }
}

/**-----------------------------------------------------------------------------
 @brief    Iterate the collection of chrono timings and display them to console,
 grouped both by percentages and raw microseconds.
*/
void disp_timing() {
  // read and sum the system timings, cast into microseconds
  vector<long> us_times{};
  for (const auto& time_spn : time_spans)  // time_spans global
    us_times.emplace_back(time_spn.time_since_epoch().count() / 1000);
  //
  auto us_sum = accumulate(cbegin(us_times), cend(us_times), 0);

  // DESIGN: use C++20 std::format when available

  // raw microseconds
  const auto   sea_t{us_times[0]};
  const auto   loc_t{us_times[1]};
  const auto   cpy_t{us_times[2]};
  const auto   srt_t{us_times[3]};
  const auto   int_t{us_times[4]};
  const auto   ord_t{us_times[5]};
  const string raws{"[sea:" + to_string(sea_t) + ", loc:" + to_string(loc_t) +
                    ", cpy:" + to_string(cpy_t) + ", srt:" + to_string(srt_t) +
                    ", int:" + to_string(int_t) + ", ord:" + to_string(ord_t) +
                    "] us"};

  // percent of whole
  const auto   s_p = rw::make_pct_str(sea_t, us_sum);
  const auto   l_p = rw::make_pct_str(loc_t, us_sum);
  const auto   c_p = rw::make_pct_str(cpy_t, us_sum);
  const auto   r_p = rw::make_pct_str(srt_t, us_sum);
  const auto   i_p = rw::make_pct_str(int_t, us_sum);
  const auto   o_p = rw::make_pct_str(ord_t, us_sum);
  const string percents{'[' + s_p + ':' + l_p + ':' + c_p + ':' + r_p + ':' +
                        i_p + ':' + o_p + "]%"};

  cout << (show_timing ? "  full search took:  " + to_string(us_sum) + " us" +
                             "  -  " + percents + " : " + raws + "\n"
                       : "\n");
  time_spans.clear();
}

/**-----------------------------------------------------------------------------
 @brief    Front end to search-loop the word@post map 'w_tp' for user's query
 terms.

 @param    w_tp [IN] The word/thread+post map to search through.
*/
void do_search(Map_words_2_posts& w_tp) {
  string srch_phrase{};
  while (true) {
    cout << "\n search:  ";
    getline(cin, srch_phrase);

    if (srch_phrase == "q")
      break;
    else {
      const auto bgn_sea = clk_time.now();

      const auto words = scrub_n_extract(srch_phrase);
      search           = assemble_words(words);  // store to global while here

      time_spans.emplace_back(clk_time.now() - bgn_sea);  // time_spans[0]: sea

      locate_terms(w_tp, words);

      if (have_results)
        disp_timing();

      disp_unfounds();
    }
  }
}

/**-----------------------------------------------------------------------------
 @brief    Clean & store the words from the string message 'post_msg' into the
 word@post map 'io_w_tp'.

           Take the input string 'post_msg', break it apart into individual
 words & clean them, then potentially store them each as a unique key into a map
 'io_w_tp'. The value for each key is a set-collection of paired
 thread_id+post_id ('thrd_id' + 'post_id')s. This effectively maps every single
 unique word to every IB post where it occurs at least once.

           Additionally, assemble all the cleaned words into another map (keyed
 by post ID) containing all the cleaned text of each post in the JSON files
 collection.

 @param    io_w_tp [IN/OUT] The word/thread+post map to store cleaned text into.
 @param    post_msg [IN] The text of the post's message.
 @param    thrd_id [IN] The thread's ID number.
 @param    post_id [IN] The post's ID number.
*/
void clean_store_words(Map_words_2_posts& io_w_tp,
                       const std::string& post_msg,
                       const uint32_t     thrd_id,
                       const uint32_t     post_id) {
  const auto words = scrub_n_extract(post_msg);

  for (const auto& word : words)
    io_w_tp[word].insert(make_pair(thrd_id, post_id));

  all_msgs[post_id] = assemble_words(words);  // store to global while here
}

/**-----------------------------------------------------------------------------
 @brief    Load the word@post map 'io_w_tp', using each JSON file located inside
 the 'all_jsons' directory.

 @param    io_w_pt [IN/OUT] The word-counting map to load.
*/
void ld_map(Map_words_2_posts& io_w_tp) {
  //
  // lambda to parse a single JSON file into the word@post map
  auto parse_json_file{[&](auto& filepath) {
    const auto thread =
        rw::get_jsonfile_dat(filepath);  // each filepath == 1 imageboard thread

    const uint32_t thrd_id{thread["threadId"].asUInt()};
    const auto     thrd_subj = thread["subject"].asString();

    // store to global while here
    thrd_subjs[thrd_id] = rw::clean_str(thrd_subj);

    // OP's subject+message text
    clean_store_words(io_w_tp,
                      string{thrd_subj + ' ' + thread["message"].asString()},
                      thrd_id, thrd_id);

    // each reply post's subject+message text
    for (const auto& post : thread["posts"])
      clean_store_words(
          io_w_tp,
          string{post["subject"].asString() + ' ' + post["message"].asString()},
          thrd_id, post["postId"].asUInt());
  }};

  // first, read in all JSON file paths from the directory
  vector<file_path> paths{};
  for (const auto& jsonfile : dir_iter{jsons_dir})  // jsons_dir{"./all_jsons/"}
    paths.push_back(jsonfile);

    // next, run file processing calling the lambda above
#if __cpp_lib_parallel_algorithm >= 201603
  for_each(par_unseq, cbegin(paths), cend(paths), parse_json_file);
#else
  for_each(cbegin(paths), cend(paths), parse_json_file);
#endif
}

/**-----------------------------------------------------------------------------
 @brief    Stream the vector 'vw_vtp' out to the full word/post/count file.

 @param    vw_vtp [IN] The vector to stream to file.
*/
void wrt_counts_file(const Vec_words_2_posts& vw_vtp) {
  auto ofs = rw::write_file_ready("word_post_counts.txt");

  for (const auto& [term, thrd_post] : vw_vtp)
    ofs << left << setw(7) << thrd_post.size() << term << '\n';
}

/**-----------------------------------------------------------------------------
 @brief    Copy word@posts map to vector & write that out to a descending-sorted
 words+counts file.

 @param    w_tp [IN] The map to write out to sorted file.
*/
void sort_wrt_counts(const Map_words_2_posts& w_tp) {
  Vec_words_2_posts vw_vtp_counts{};
  cp_map_2_vec(w_tp, vw_vtp_counts);
  sort_counts(vw_vtp_counts, true);
  wrt_counts_file(vw_vtp_counts);
}

/**-----------------------------------------------------------------------------
 @brief    Write the 'all_posts' map containing all the cleaned text of all the
 posts out to file.
*/
void wrt_all_posts() {
  auto ofs = rw::write_file_ready("all_posts.txt");

  for (const auto& [post_id, post_msg] : all_msgs)
    ofs << post_id << '\n' << post_msg << "\n\n";
}

/**-----------------------------------------------------------------------------
 @brief    The program entry point.

 @returns  The program's exit status.
*/
int main() {
  auto bgn_ldwrt = clk_time.now();

  // load the full word/thread_id+post_id (word@post) map from JSON files
  Map_words_2_posts w_tp{};
  ld_map(w_tp);

  // DESIGN: create a task-system, and then just queue the two file writes?
  if (write_files) {
    wrt_all_posts();        // write all post's cleaned plaintext out to file
    sort_wrt_counts(w_tp);  // write sorted word counts out to file
  }

  cout << "time to process local /" + board_nm + "/ JSON data:  "
       << duration_cast<milliseconds>(clk_time.now() - bgn_ldwrt).count()
       << " ms\n";

  do_search(w_tp);  // search-loop the word@post map
}

// Copyright (2021)
// License (MIT) https://opensource.org/licenses/MIT
